@@ -1,7 +1,7 @@
 use anyhow::Result;
 use clap::{App, Arg};
 use hgrep::grep::BufReadExt;
-use hgrep::printer::{BatPrinter, Printer};
+use hgrep::printer::{PrinterOptions, TextWrapMode};
 use std::cmp;
 use std::env;
 use std::io;
@@ -10,7 +10,19 @@ use std::process;
 #[cfg(feature = "ripgrep")]
 use hgrep::ripgrep;
 
+#[cfg(feature = "bat-printer")]
+use hgrep::bat::BatPrinter;
+
+#[cfg(feature = "syntect-printer")]
+use hgrep::syntect::SyntectPrinter;
+
 fn cli<'a>() -> App<'a> {
+    #[cfg(feature = "bat-printer")]
+    const DEFAULT_PRINTER: &str = "bat";
+
+    #[cfg(all(not(feature = "bat-printer"), feature = "syntect-printer"))]
+    const DEFAULT_PRINTER: &str = "syntect";
+
     let app = App::new("hgrep")
         .version(env!("CARGO_PKG_VERSION"))
         .about(
@@ -41,12 +53,12 @@ fn cli<'a>() -> App<'a> {
             Arg::new("no-grid")
                 .short('G')
                 .long("no-grid")
-                .about("Remove border lines for more compact output"),
+                .about("Remove borderlines for more compact output"),
         )
         .arg(
             Arg::new("grid")
                 .long("grid")
-                .about("Add border lines to output. This flag is an opposite of --no-grid"),
+                .about("Add borderlines to output. This flag is an opposite of --no-grid"),
         )
         .arg(
             Arg::new("tab")
@@ -68,14 +80,58 @@ fn cli<'a>() -> App<'a> {
                 .long("list-themes")
                 .about("List all theme names available for --theme option"),
         )
+        .arg(
+            Arg::new("printer")
+                .short('p')
+                .long("printer")
+                .value_name("PRINTER")
+                .default_value(DEFAULT_PRINTER)
+                .about("Printer to print the match results. 'bat' or 'syntect' is available"),
+        )
+        .arg(
+            Arg::new("term-width")
+                .long("term-width")
+                .takes_value(true)
+                .value_name("NUM")
+                .about("Width (number of characters) of terminal window"),
+        ).arg(
+            Arg::new("no-wrap")
+                .long("no-wrap")
+                .about("DEPRECATED: The same as `--wrap never`. This flag will be removed at v0.2.0")
+        ).arg(
+            Arg::new("wrap")
+                .long("wrap")
+                .takes_value(true)
+                .value_name("MODE")
+                .default_value("char")
+                .possible_values(["char", "never"])
+                .case_insensitive(true)
+                .about("Text-wrapping mode. 'char' enables character-wise text-wrapping. 'never' disables text-wrapping")
+        )
         .arg(
             Arg::new("generate-completion-script")
                 .long("generate-completion-script")
                 .takes_value(true)
                 .value_name("SHELL")
-                .about("Print completion script for SHELL to stdout. SHELL must be one of 'bash', 'zsh', 'powershell', 'fish', or 'elvish'"),
+                .possible_values(["bash", "zsh", "powershell", "fish", "elvish"])
+                .case_insensitive(true)
+                .about("Print completion script for SHELL to stdout"),
         );
 
+    #[cfg(feature = "bat-printer")]
+    let app = app.arg(
+        Arg::new("custom-assets")
+            .long("custom-assets")
+            .about("Load bat's custom assets. Note that this flag may not work with some version of `bat` command. This flag is only for bat printer"),
+    );
+
+    #[cfg(feature = "syntect-printer")]
+    let app = app.arg(
+        Arg::new("background")
+            .long("background")
+            .about("Paint background colors. This flag is only for syntect printer"),
+    );
+
     #[cfg(feature = "ripgrep")]
     let app = app
             .about(
@@ -230,22 +286,34 @@ fn cli<'a>() -> App<'a> {
     app
 }
 
-fn generate_completion_script(shell: &str) -> Result<()> {
+fn generate_completion_script(shell: &str) {
     use clap_generate::generate;
     use clap_generate::generators::*;
 
     let mut app = cli();
     let stdout = io::stdout();
     let mut stdout = stdout.lock();
-    match shell.to_lowercase().as_str() {
-        "bash" => generate(Bash, &mut app, "hgrep", &mut stdout),
-        "zsh" => generate(Zsh, &mut app, "hgrep", &mut stdout),
-        "powershell" => generate(PowerShell, &mut app, "hgrep", &mut stdout),
-        "fish" => generate(Fish, &mut app, "hgrep", &mut stdout),
-        "elvish" => generate(Elvish, &mut app, "hgrep", &mut stdout),
-        s => anyhow::bail!("Unknown shell '{}'. Supported shells are 'bash', 'zsh', 'powershell', 'fish', or 'elvish'", s),
+    if shell.eq_ignore_ascii_case("bash") {
+        generate(Bash, &mut app, "hgrep", &mut stdout)
+    } else if shell.eq_ignore_ascii_case("zsh") {
+        generate(Zsh, &mut app, "hgrep", &mut stdout)
+    } else if shell.eq_ignore_ascii_case("powershell") {
+        generate(PowerShell, &mut app, "hgrep", &mut stdout)
+    } else if shell.eq_ignore_ascii_case("fish") {
+        generate(Fish, &mut app, "hgrep", &mut stdout)
+    } else if shell.eq_ignore_ascii_case("elvish") {
+        generate(Elvish, &mut app, "hgrep", &mut stdout)
+    } else {
+        unreachable!() // SHELL argument was validated by clap
     }
-    Ok(())
+}
+
+#[derive(Clone, Copy, PartialEq, Eq)]
+enum PrinterKind {
+    #[cfg(feature = "bat-printer")]
+    Bat,
+    #[cfg(feature = "syntect-printer")]
+    Syntect,
 }
 
 fn app() -> Result<bool> {
@@ -253,16 +321,42 @@ fn app() -> Result<bool> {
 
     let matches = cli().get_matches();
     if let Some(shell) = matches.value_of("generate-completion-script") {
-        generate_completion_script(shell)?;
+        generate_completion_script(shell);
         return Ok(true);
     }
 
-    let mut printer = BatPrinter::new();
+    #[allow(unused_variables)] // printer_kind is unused when syntect-printer is disabled for now
+    let printer_kind = match matches.value_of("printer").unwrap() {
+        #[cfg(feature = "bat-printer")]
+        "bat" => PrinterKind::Bat,
+        #[cfg(not(feature = "bat-printer"))]
+        "bat" => anyhow::bail!("--printer bat is not available because 'bat-printer' feature was disabled at compilation"),
+        #[cfg(feature = "syntect-printer")]
+        "syntect" => PrinterKind::Syntect,
+        #[cfg(not(feature = "syntect-printer"))]
+        "syntect" => anyhow::bail!("--printer syntect is not available because 'syntect-printer' feature was disabled at compilation"),
+        p => anyhow::bail!(
+            "Unknown printer '{}', at --printer option. It must be one of 'bat' or 'syntect'",
+            p
+        ),
+    };
+
     if matches.is_present("list-themes") {
-        for theme in printer.themes() {
-            println!("{}", theme);
+        #[cfg(feature = "syntect-printer")]
+        if printer_kind == PrinterKind::Syntect {
+            hgrep::syntect::list_themes(io::stdout().lock())?;
+            return Ok(true);
         }
-        return Ok(true);
+
+        #[cfg(feature = "bat-printer")]
+        if printer_kind == PrinterKind::Bat {
+            for theme in BatPrinter::new(PrinterOptions::default()).themes() {
+                println!("{}", theme);
+            }
+            return Ok(true);
+        }
+
+        unreachable!();
     }
 
     let min_context = matches
@@ -277,26 +371,70 @@ fn app() -> Result<bool> {
         .context("could not parse \"max-context\" option value as unsigned integer")?;
     let max_context = cmp::max(min_context, max_context);
 
-    let tab_width = matches
-        .value_of("tab")
-        .unwrap()
-        .parse()
-        .context("could not parse \"tab\" option value as unsigned integer")?;
-    printer.tab_width(tab_width);
+    let mut printer_opts = PrinterOptions::default();
+    if let Some(width) = matches.value_of("tab") {
+        printer_opts.tab_width = width
+            .parse()
+            .context("could not parse \"tab\" option value as unsigned integer")?;
+    }
 
     let theme_env = env::var("BAT_THEME").ok();
     if let Some(var) = &theme_env {
-        printer.theme(var);
+        printer_opts.theme = Some(var);
     }
     if let Some(theme) = matches.value_of("theme") {
-        printer.theme(theme);
+        printer_opts.theme = Some(theme);
     }
 
+    let is_grid = matches.is_present("grid");
     if let Ok("plain" | "header" | "numbers") = env::var("BAT_STYLE").as_ref().map(String::as_str) {
-        printer.no_grid();
+        if !is_grid {
+            printer_opts.grid = false;
+        }
     }
-    if matches.is_present("no-grid") && !matches.is_present("grid") {
-        printer.no_grid();
+    if matches.is_present("no-grid") && !is_grid {
+        printer_opts.grid = false;
+    }
+
+    if let Some(width) = matches.value_of("term-width") {
+        let width = width
+            .parse()
+            .context("could not parse \"term-width\" option value as unsigned integer")?;
+        printer_opts.term_width = width;
+        if width < 10 {
+            anyhow::bail!("Too small value at --term-width option ({} < 10)", width);
+        }
+    }
+
+    if let Some(mode) = matches.value_of("wrap") {
+        if mode.eq_ignore_ascii_case("never") {
+            printer_opts.text_wrap = TextWrapMode::Never;
+        } else if mode.eq_ignore_ascii_case("char") {
+            printer_opts.text_wrap = TextWrapMode::Char;
+        } else {
+            unreachable!(); // Option value was validated by clap
+        }
+    }
+    if matches.is_present("no-wrap") {
+        printer_opts.text_wrap = TextWrapMode::Never;
+    }
+
+    #[cfg(feature = "syntect-printer")]
+    if matches.is_present("background") {
+        printer_opts.background_color = true;
+        #[cfg(feature = "bat-printer")]
+        if printer_kind == PrinterKind::Bat {
+            anyhow::bail!("--background flag is only available for syntect printer since bat does not support painting background colors");
+        }
+    }
+
+    #[cfg(feature = "bat-printer")]
+    if matches.is_present("custom-assets") {
+        printer_opts.custom_assets = true;
+        #[cfg(feature = "syntect-printer")]
+        if printer_kind == PrinterKind::Syntect {
+            anyhow::bail!("--custom-assets flag is only available for bat printer");
+        }
     }
 
     #[cfg(feature = "ripgrep")]
@@ -362,28 +500,55 @@ fn app() -> Result<bool> {
             config.types_not(types_not);
         }
 
-        if let Some(paths) = paths {
-            return ripgrep::grep(printer, pattern, paths, config);
-        } else {
-            let cwd = env::current_dir()?;
-            let paths = std::iter::once(cwd.as_os_str());
+        #[cfg(feature = "syntect-printer")]
+        if printer_kind == PrinterKind::Syntect {
+            let printer = SyntectPrinter::with_stdout(printer_opts)?;
             return ripgrep::grep(printer, pattern, paths, config);
         }
+
+        #[cfg(feature = "bat-printer")]
+        if printer_kind == PrinterKind::Bat {
+            let printer = std::sync::Mutex::new(BatPrinter::new(printer_opts));
+            return ripgrep::grep(printer, pattern, paths, config);
+        }
+
+        unreachable!();
     }
 
-    let mut found = false;
-    // XXX: io::stdin().lock() is not available since bat's implementation internally takes lock of stdin
-    // *even if* it does not use stdin.
-    // https://github.com/sharkdp/bat/issues/1902
-    for f in io::BufReader::new(io::stdin())
-        .grep_lines()
-        .chunks_per_file(min_context, max_context)
-    {
-        printer.print(f?)?;
-        found = true;
+    #[cfg(feature = "syntect-printer")]
+    if printer_kind == PrinterKind::Syntect {
+        use hgrep::printer::Printer;
+        use rayon::prelude::*;
+        let printer = SyntectPrinter::with_stdout(printer_opts)?;
+        return io::BufReader::new(io::stdin())
+            .grep_lines()
+            .chunks_per_file(min_context, max_context)
+            .par_bridge()
+            .map(|file| {
+                printer.print(file?)?;
+                Ok(true)
+            })
+            .try_reduce(|| false, |a, b| Ok(a || b));
     }
 
-    Ok(found)
+    #[cfg(feature = "bat-printer")]
+    if printer_kind == PrinterKind::Bat {
+        let mut found = false;
+        let printer = BatPrinter::new(printer_opts);
+        // XXX: io::stdin().lock() is not available since bat's implementation internally takes lock of stdin
+        // *even if* it does not use stdin.
+        // https://github.com/sharkdp/bat/issues/1902
+        for f in io::BufReader::new(io::stdin())
+            .grep_lines()
+            .chunks_per_file(min_context, max_context)
+        {
+            printer.print(f?)?;
+            found = true;
+        }
+        return Ok(found);
+    }
+
+    unreachable!();
 }
 
 fn main() {
@@ -396,7 +561,8 @@ fn main() {
         Ok(true) => 0,
         Ok(false) => 1,
         Err(err) => {
-            eprintln!("{}", err);
+            let red = ansi_term::Colour::Red.bold();
+            eprintln!("{} {}", red.paint("error:"), err);
             2
         }
     };
