@@ -3,6 +3,12 @@ use anyhow::Result;
 use std::env;
 use term::terminfo::TermInfo;
 
+#[derive(Clone, Copy, PartialEq, Eq)]
+pub enum TextWrapMode {
+    Char,
+    Never,
+}
+
 #[derive(Clone, Copy, PartialEq, Eq)]
 pub enum TermColorSupport {
     True,
@@ -14,7 +20,7 @@ impl TermColorSupport {
     fn detect() -> Self {
         if env::var("COLORTERM")
             .ok()
-            .map(|v| v.eq_ignore_ascii_case("truecolor"))
+            .map(|v| v.eq_ignore_ascii_case("truecolor") || v.eq_ignore_ascii_case("24bit"))
             .unwrap_or(false)
         {
             return TermColorSupport::True;
@@ -41,7 +47,7 @@ pub struct PrinterOptions<'main> {
     pub color_support: TermColorSupport,
     pub term_width: u16,
     pub custom_assets: bool,
-    pub text_wrap: bool,
+    pub text_wrap: TextWrapMode,
 }
 
 impl<'main> Default for PrinterOptions<'main> {
@@ -54,8 +60,8 @@ impl<'main> Default for PrinterOptions<'main> {
             background_color: false,
             color_support: TermColorSupport::detect(),
             custom_assets: false,
-            term_width: terminal_size().map(|(Width(w), _)| w).unwrap_or(80),
-            text_wrap: true,
+            term_width: terminal_size().map(|(Width(w), _)| w).unwrap_or(80), // Note: `tput` returns 80 when tty is not found
+            text_wrap: TextWrapMode::Char,
         }
     }
 }
